/*
 * Copyright (c) 2013 JST DEOS R&D Center
 */
/*
 * rtkl_collect.c:
 *	rtkl-collect is an user-land support program of RootkitLibra (RTKL)
 *	application which is intended to collect untrsuted view of file-system
 *	access.  Untrsuted view consist of meta data (inode number, file size,
 *	and number of files in the directory) of files and directories seen
 *	by user-land context of User Guest OS, collected through some system
 *	calls (open, close, fstat, lstat, and getdents).  Since collected
 *	data will be parsed and analized by observer ('dsysmon' program
 *	running under Watcher Guest OS), rtkl-collect simply send collected
 *	data, argument and result of each system-calls, to the direct
 *	cross-feed channel.  
 *
 *	Typical syntax to start rtkl-collect is,
 *
 *	$ ./rtkl-collect -d 1 -r ./mnt
 *
 *	where "-d" option specifies directory herarchy depth to be traversed,
 *	"-r" option specifies starting directory of the meta-data collection.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/queue.h>
#include <libgen.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

//#include <dsm_api.h>


struct linux_dirent {
	long	       d_ino;
	off_t	       d_off;
	unsigned short d_reclen;
	char	       d_name[];
};

struct directory {
	char* name;
	LIST_ENTRY(directory) next;
};

LIST_HEAD(list_dirs, directory);


char*	version_string = "0.0.1";
char*	program_name   = NULL;
bool	is_debug       = false;
char*	root_directory = NULL;
int32_t	depth	       = 0;
int32_t send_delay     = 0;
int32_t	max_send_size;


#define DEFAULT_ROOT_DIR	"/"

#define BUF_SIZE	3072

#define MAX_BUF_SIZE	4096

#define DEBUG_PRINTF(format, ...)					\
	do {								\
		if (is_debug == true) {					\
			fprintf (stderr, format, ## __VA_ARGS__);	\
			fsync (2);					\
		}							\
	} while (0)

#define __UNUSED	__attribute__((unused))

#define STRING_EQL(s1, s2)						\
	((strncmp (s1, s2, strlen (s1)) == 0) &&			\
	 (strlen (s1) == strlen (s2)))

#define MEMCPY(to, from, size)			\
	do {					\
		memcpy (to, from, size);	\
		to += size;			\
	} while (0)

#define STORE_INT32(p, val)			\
	do {					\
		p[3] = (val >> 24) & 0xff;	\
		p[2] = (val >> 16) & 0xff;	\
		p[1] = (val >>  8) & 0xff;	\
		p[0] = (val >>  0) & 0xff;	\
		p += 4;				\
	} while (0)


/* Send data to Watcher OS from User OS. */
static void
rtkl_dxfeed_send (char* data, int32_t len)
{
	int n;

	if (max_send_size < len) {
		fprintf (stderr, "E: writing data is too large (%d > %d)\n",
			 len, max_send_size);
		exit (EXIT_FAILURE);
	}

#if 1
	if ((n = dsm_dxfeed_write (data, len)) < 0) {
		fprintf (stderr, "E: dsm_dxfeed_write (%d)\n", n);
		exit (EXIT_FAILURE);
	}
#endif

	if (send_delay > 0) {
		usleep (send_delay * 1000);
	}

	return ;
}

static void*
rtkl_malloc (size_t size)
{
	void* p;

	if ((p = malloc (size)) == NULL && size != 0) {
		perror ("malloc");
		exit (EXIT_FAILURE);
	}

	return p;
}

static void
rtkl_free (void* p)
{
	if (p != NULL) {
		free (p);
	}
}

/* Hook open systemcall. Send arguments and return value to Watcher OS. */
static int
rtkl_open (char* filename, int flags)
{
	int	n, e;
	char	data[MAX_BUF_SIZE];
	char*	p	= data;
	int	str_len = 0;

printf("***** open *****\n");
fsync(1);
	n = open (filename, flags);
	e = errno;

	STORE_INT32 (p, SYS_open);
	str_len = strlen (filename);
	STORE_INT32 (p, str_len);
	MEMCPY (p, filename, str_len);
	STORE_INT32 (p, flags);
	STORE_INT32 (p, n);

	DEBUG_PRINTF ("open> "
		      "call = %d, filename = %s (%d) flags = %d, ret = %d\n",
		      SYS_open, filename, strlen (filename), flags, n);
	rtkl_dxfeed_send (data, (p - data));

	errno = e;
	return n;
}

/* Hook close systemcall. Send arguments and return value to Watcher OS. */
static int
rtkl_close (int fd)
{
	int n, e;
	char	data[MAX_BUF_SIZE];
	char*	p	= data;

printf("***** close *****\n");
fsync(1);
	n = close (fd);
	e = errno;

	STORE_INT32 (p, SYS_close);
	STORE_INT32 (p, fd);
	STORE_INT32 (p, n);

	DEBUG_PRINTF ("close> call = %d, fd = %d, ret = %d\n",
		      SYS_close, fd, n);
	rtkl_dxfeed_send (data, (p - data));

	errno = e;
	return n;
}

/* Hook fstat systemcall. Send arguments and return value to Watcher OS. */
static int
rtkl_fstat (int fd, struct stat* buf)
{
	int	n, e;
	char	data[MAX_BUF_SIZE];
	char*	p   = data;

printf("***** fstat *****\n");
fsync(1);
	n = fstat (fd, buf);
	e = errno;

	STORE_INT32 (p, SYS_fstat);
	STORE_INT32 (p, fd);
	MEMCPY (p, (char *) buf, sizeof (struct stat));
	STORE_INT32 (p, n);

	DEBUG_PRINTF ("fstat> "
		      "call = %d, fd = %d, ino = %ld, size = %lxd, ret = %d\n",
		      SYS_fstat, fd, buf->st_ino, buf->st_size, n);
	rtkl_dxfeed_send (data, (p - data));

	errno = e;
	return n;
}

/* Hook lstat systemcall. Send arguments and return value to Watcher OS. */
static int
rtkl_lstat (char* filename, struct stat* buf)
{
	int	n, e;
	char	data[MAX_BUF_SIZE];
	char*	p	= data;
	int	str_len = 0;

printf("***** lstat *****\n");
fsync(1);
	n = lstat (filename, buf);
	e = errno;

	STORE_INT32 (p, SYS_lstat);
	str_len = strlen (filename);
	STORE_INT32 (p, str_len);
	MEMCPY (p, filename, str_len);
	MEMCPY (p, (char*) buf, sizeof (struct stat));
	STORE_INT32 (p, n);

	DEBUG_PRINTF ("lstat> "
		      "call = %d, name = %s (%d), ""ino = %ld, size = %ld, "
		      "ret = %d\n",
		      SYS_lstat, filename, strlen (filename), buf->st_ino,
		      buf->st_size, n);
	rtkl_dxfeed_send (data, (p - data));

	errno = e;
	return n;
}

/* Hook syscall (getdents) systemcall. Send arguments and return value
 * to Watcher OS. */
static int
rtkl_syscall(int sysno, int fd, char* dirp, int count)
{
	int	n, e;
	char	data[MAX_BUF_SIZE];
	char*	p	= data;

printf("***** syscall *****\n");
fsync(1);
	n = syscall (sysno, fd, dirp, count);
	e = errno;

printf("***** syscall ***** 0 p = 0x%08x\n", p);
fsync(1);
	STORE_INT32 (p, sysno);
printf("***** syscall ***** 1 sysno %d  p = 0x%08x %02x %02x %02x %02x\n", sysno, p, p[-4], p[-3], p[-2], p[-1]);
fsync(1);
	STORE_INT32 (p, fd);
printf("***** syscall ***** 2 fd %d  p = 0x%08x %02x %02x %02x %02x\n", fd, p, p[-4], p[-3], p[-2], p[-1]);
fsync(1);
	STORE_INT32 (p, count);
printf("***** syscall ***** 3 count %d  p = 0x%08x %02x %02x %02x %02x\n", count, p, p[-4], p[-3], p[-2], p[-1]);
fsync(1);
	MEMCPY (p, dirp, count);
printf("***** syscall ***** 4 dirp p = 0x%08x %02x %02x %02x %02x\n", p, p[-4], p[-3], p[-2], p[-1]);
fsync(1);
	STORE_INT32 (p, count);
printf("***** syscall ***** 5 count %d  p = 0x%08x %02x %02x %02x %02x\n", count, p, p[-4], p[-3], p[-2], p[-1]);
fsync(1);
	STORE_INT32 (p, n);
printf("***** syscall ***** 6 n %d  p = 0x%08x %02x %02x %02x %02x\n", n, p, p[-4], p[-3], p[-2], p[-1]);
fsync(1);

int i;
printf("data = ");
for(i = 0; i < 16; i++) {
    printf("%02x ", (unsigned char)data[i]);
}
printf("\n");
printf("data[12+%d] = ", count);
for(i = 0; i < 8; i++) {
    printf("%02x ", (unsigned char)data[12 + count + i]);
}
printf("\n");
printf("dirp = ");
for(i = 0; i < 16; i++) {
    printf("%02x ", (unsigned char)dirp[i]);
}
printf("\n");
fsync(1);
	printf ("getdents> "
		      "call = %d, fd = %d, strlen = %d count = %d ret = %d\n",
		      sysno, fd, count, count, n);
fsync(1);

	DEBUG_PRINTF ("getdents> "
		      "call = %d, fd = %d, strlen = %d count = %d ret = %d\n",
		      sysno, fd, count, count, n);
	rtkl_dxfeed_send (data, (p - data));

	errno = e;
	return n;
}

/* Convert number string to integer. */
static int32_t
string_to_int (char* str, int32_t base)
{
	int   num;
	char* ep;
	long  val;

	errno = 0;
	val = strtol (str, &ep, base);
	if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN)) ||
	    (errno != 0 && val == 0)) {
		perror ("strtol");
		return -1;
	}

	if (ep == optarg) {
		fprintf (stderr, "No digits were found.\n");
		return -1;
	}

	return (num = val);
}

static void
version_show (FILE* stream)
{
	fprintf (stream, "Version %s (%s %s)\n",
		 version_string, __DATE__, __TIME__);
}

static void
usage_show (FILE* stream)
{
	fprintf (stream,
		 "Usage: %s\n"
		 "Options:\n"
		 "	-D\n"
		 "		Dispaly debug purpose messages too.\n"
		 "	-d <depth>\n"
		 "		Specify a depth of the searching.\n"
		 "	-r <directory>\n"
		 "		Specify a searching top directory.\n"
		 "	-h\n"
		 "		Show this message and exit.\n"
		 "	-s <delay>\n"
		 "		Specify a delay (mili sec) that waits after "
		 "calling systemcall.\n"
		 "	-v\n"
		 "		Print version and exit.\n",
		 program_name);
}

/* Get file size of the regular file. */
static u_int32_t
filesize_get (char* base_dir, char* file_name)
{
	int32_t		fd, len, size = 0;
	char*		name;
	struct stat	st;

	len  = strlen (base_dir) + strlen (file_name);
	name = (char*) rtkl_malloc (len + 1);
	if (snprintf (name, len + 1, "%s%s", base_dir, file_name) != len) {
		perror ("snprintf");
		goto file_free;
	}

	if ((fd = rtkl_open (name, O_RDONLY)) < 0) {
		perror ("open");
		goto fd_close;
	}

	if (rtkl_fstat (fd, &st) < 0) {
		perror ("fstat");
		goto fd_close;
	}
	size = st.st_size;

fd_close:
	if (rtkl_close (fd) < 0) {
		perror ("close");
	}
file_free:
	rtkl_free (name);

	return size;
}

/* Get file size of the symbolic file. */
static u_int32_t
linksize_get (char* base_dir, char* file_name)
{
	int32_t		len, size = 0;
	char*		name;
	struct stat	st;

	len  = strlen (base_dir) + strlen (file_name);
	name = (char*) rtkl_malloc (len + 1);
	if (snprintf (name, len + 1, "%s%s", base_dir, file_name) != len) {
		perror ("snprintf");
		goto file_free;
	}

	if (rtkl_lstat (name, &st) < 0) {
		perror ("stat");
		goto file_free;
	}
	size = st.st_size;

file_free:
	rtkl_free (name);
	return size;
}

/* Add directory to the list `dirs'. */
static void
directory_add (struct list_dirs* dirs, char* base_dir, char* name)
{
	int len;
	struct directory* d = NULL;

	d	= (struct directory*) rtkl_malloc (sizeof (struct directory));
	len	= strlen (base_dir) + strlen (name) + strlen ("/");
	d->name = (char*) rtkl_malloc (len + 1);

	if (snprintf (d->name, len + 1, "%s%s/", base_dir, name) != len) {
		perror ("snprintf");
		return ;
	}
	LIST_INSERT_HEAD(dirs, d, next);
}

/* find all files in the directory `dir'. */
static void
find_directory_files (char* dir, int32_t cur_depth)
{
	int			 fd, nr, pos, size;
	int			 files = 0;
	char			 d_type;
	char			 buf[BUF_SIZE];
	struct linux_dirent	*entry;
	struct stat st;
	struct list_dirs dirs;

	LIST_INIT(&dirs);

	if ((fd = rtkl_open (dir, O_RDONLY)) < 0) {
		perror ("open");
		return ;
	}

	while ((depth == 0) || (depth != cur_depth)) {
		if ((nr = rtkl_syscall (SYS_getdents, fd, buf, BUF_SIZE)) < 0) {
			perror ("getdents");
			goto fd_close;
		}

		/* read all directory entries. */
		if (nr == 0) {
			break;
		}

		for (pos=0; pos<nr;) {
			entry = (struct linux_dirent*) (buf + pos);
			d_type = *(buf + pos + entry->d_reclen -1);
			pos += entry->d_reclen;

			if (STRING_EQL (".",  entry->d_name) ||
			    STRING_EQL ("..", entry->d_name)) {
				continue;
			}

			switch (d_type) {
			case DT_REG:
				size = filesize_get (dir, entry->d_name);
				files++;
				break;
			case DT_LNK:
				size = linksize_get (dir, entry->d_name);
				files++;
				break;
			case DT_DIR:
				directory_add (&dirs, dir, entry->d_name);
				files++;
				break;
			default:
				break;
			}
		}
	}

	if (rtkl_fstat (fd, &st) < 0) {
		perror ("stat");
	}

fd_close:
	if (rtkl_close (fd) < 0) {
		perror ("close");
	}

	struct directory* dp;
	struct directory* dp_next;

	for (dp = dirs.lh_first; dp != NULL; dp = dp_next) {
		dp_next = LIST_NEXT (dp, next);
		find_directory_files (dp->name, cur_depth + 1);
		rtkl_free (dp->name);
		LIST_REMOVE(dp, next);
		rtkl_free (dp);
	}

	return ;
}

static char*
rootdir_set (char* dir)
{
	int len = strlen (dir);
	char* d;

	d = rtkl_malloc (len + 2);
	snprintf (d, len + 2, "%s/", dir);

	return d;
}

int
main (int argc, char* argv[])
{
	static struct option opts[] = {
		{"--debug",		required_argument,	NULL,	'D'},
		{"--depth",		required_argument,	NULL,	'd'},
		{"--help",		no_argument,		NULL,	'h'},
		{"--root-dir",		required_argument,	NULL,	'r'},
		{"--send-delay",	required_argument,	NULL,	's'},
		{"--version",		no_argument,		NULL,	'v'},
		{ NULL,			0,			NULL,	0},
	};

	program_name = basename (argv[0]);

	int c;
	while ((c = getopt_long (argc, argv, "Dd:hr:s:v", opts, NULL)) != EOF) {
		switch (c) {
		case 'D':
			is_debug = true;
			break;
		case 'd':
			depth = string_to_int (optarg, 10);
			break;
		case 'r':
			if (opendir (optarg) == NULL) {
				perror ("opendir");
				exit (EXIT_FAILURE);
			}
			root_directory = rootdir_set (optarg);
			break;
		case 's':
			send_delay = string_to_int (optarg, 10);
			break;
		case 'h':
			usage_show (stdout);
			exit (EXIT_SUCCESS);
		case 'v':
			version_show (stdout);
			exit (EXIT_SUCCESS);
		default:
			usage_show (stderr);
			exit (EXIT_FAILURE);
		}
	}

	max_send_size = dsm_dxfeed_maxsize();
	find_directory_files ((root_directory == NULL) ? DEFAULT_ROOT_DIR :
			      root_directory, 0);

	return EXIT_SUCCESS;
}

/* findfs.c ends here. */
